import json
import os
import random
from collections import deque
from typing import List, Dict, Tuple

class WumpusAgent:
    def __init__(self, config_file: str):
        self.load_config(config_file)
        self.map = []
        self.position = (0, 0)
        self.climbing_gear = False
        self.response_ratings = deque(maxlen=1000)
        self.env_points_criteria = self.get_points_criteria()

    def load_config(self, config_file: str):
        with open(config_file, 'r') as file:
            config = json.load(file)
            self.agent_name = config["agent"]
            self.env = config["env"]
            self.password = config["pwd"]
            self.url = config["url"]

    def get_points_criteria(self):
        criteria = {
            "env-1.json": [(1.9, 0), (1.5, 20), (float('-inf'), 30)],
            "env-2.json": [(4.0, 0), (3.0, 30), (float('-inf'), 40)],
            "env-3.json": [(4.0, 0), (3.0, 40), (float('-inf'), 50)],
            "env-4.json": [(5.0, 0), (4.0, 45), (float('-inf'), 55)],
            "env-5.json": [(5.0, 0), (4.0, 50), (3.0, 60), (float('-inf'), 70)],
            "env-6.json": [(20.0, 0), (15.0, 50), (13.0, 75), (float('-inf'), 80)],
        }
        return criteria.get(self.env, [])

    def parse_map(self, map_str: str) -> None:
        self.map = [list(row) for row in map_str.splitlines()]
        print("Parsed Map:", self.map)

    def set_current_position(self, cell: str) -> None:
        for x in range(len(self.map)):
            for y in range(len(self.map[0])):
                if self.map[x][y] == cell:
                    self.position = (x, y)
                    print("Current Position set to:", self.position)
                    return
        raise ValueError("Current cell not found in the map.")

    def find_path_to_cave(self) -> List[str]:
        directions = {"north": (-1, 0), "south": (1, 0), "east": (0, 1), "west": (0, -1)}
        queue = deque([(self.position, [])])
        visited = set([self.position])
        obstacles = {"B"}
        path_to_cave = []

        while queue:
            (x, y), path = queue.popleft()
            if self.map[x][y] == "W":
                path_to_cave = path
                break
            
            for direction, (dx, dy) in directions.items():
                nx, ny = x + dx, y + dy
                if 0 <= nx < len(self.map) and 0 <= ny < len(self.map[0]) and (nx, ny) not in visited:
                    if self.map[nx][ny] not in obstacles:
                        visited.add((nx, ny))
                        queue.append(((nx, ny), path + [f"GO {direction}"]))
        
        return path_to_cave

    def calculate_success_chance(self, path: List[str]) -> float:
        if not path:
            return 0.0
        
        obstacles = {"B": 0,"C": 0, "R": 0, "M": 0}
        x, y = self.position
        
        for move in path:
            direction = move.split()[1]
            if direction == "north":
                x -= 1
            elif direction == "south":
                x += 1
            elif direction == "east":
                y += 1
            elif direction == "west":
                y -= 1
            
            if 0 <= x < len(self.map) and 0 <= y < len(self.map[0]):
                cell = self.map[x][y]
                if cell in obstacles:
                    obstacles[cell] += 1

        total_obstacles = sum(obstacles.values())
        success_chance = max(0.3, 1.0 - (total_obstacles / (len(path) + total_obstacles)))

        success_chance += random.uniform(-0.1, 0.1)  # Add random variation
        return round(success_chance, 2)

    def calculate_expected_time(self, path: List[str]) -> float:
        time_factor = 1.0
        for step in path:
            x, y = self.position
            if "B" in self.map[x][y]:  # Extra time for dangerous cells
                time_factor += 0.3
        expected_time = len(path) * time_factor
        expected_time += random.uniform(-0.5, 0.5)  # Add random variation
        return round(expected_time, 2)

    def calculate_rating(self, success_chance: float, expected_time: float, max_time: float) -> float:
        return round(success_chance * expected_time + (1 - success_chance) * max_time, 2)

    def calculate_points(self, rating: float) -> int:
        for threshold, points in self.env_points_criteria:
            if rating <= threshold:
                return points
        return 0

    def interact_with_server(self, observations: Dict) -> Dict:
        self.parse_map(observations["map"])
        self.climbing_gear = "climbing-gear" in observations["initial-equipment"]
        self.set_current_position(observations["observations"]["current-cell"])

        path = self.find_path_to_cave()
        success_chance = self.calculate_success_chance(path)
        expected_time = self.calculate_expected_time(path)
        max_time = observations["max-time"]

        rating = self.calculate_rating(success_chance, expected_time, max_time)
        self.response_ratings.append(rating)
        points = self.calculate_points(rating)

        response = {
            "actions": path,
            "success-chance": success_chance,
            "expected-time": expected_time,
            "rating": rating,
            "points": points
        }
        return response

    def get_average_rating(self) -> float:
        return round(sum(self.response_ratings) / len(self.response_ratings), 2) if self.response_ratings else float('inf')

# Example use
config_file = "env-4.json"
agent = WumpusAgent(config_file)
observations = {
    "initial-equipment": ["climbing-gear"],
    "map": "BBRRB\nBBCRB\nWBRRR\nBBWRB\nBBWRR",
    "max-time": 6,
    "observations": {"cell-west": "R","current-cell": "C"}
}

response = agent.interact_with_server(observations)
print("Agent Response:", response)
average_rating = agent.get_average_rating()
print("Average Rating of Last 1000 Responses:", average_rating)
